# imports
import sys
import re
import os
import html5lib
import threading
from html5lib import treebuilders
from html5lib import treewalkers
from xml.dom.ext.reader import Sax2
from xml.dom.NodeFilter import NodeFilter

# Global variables
# This is a global variable used for the indexing of elements
fileType = ""
indexId = 0 
maxDepth = 0
elementTypes1 = {"_":"_"} #this is a list of all labels we see in each file
elementTypes2 = {"_":"_"}
Mglobal = set()
matchAttrG = True

# shared methods

class Node:
	def __init__(self, label, type, parent, position, attributes):
		self.label = label.lower()
		self.type = type.lower()
		self.parent = parent
		self.position = position
		self.identifier = -1
		self.depth = -1
		self.originalPosition = -1
		self.children = []
		self.attrs = attributes #dict of attributes


def printNodeH (node, count):
	#for i in range(count):
	#	sys.stdout.write("--")
	if (node.type != "text" and len(node.label) > 0 and node.label != "no_label"):
		try: 
			sys.stdout.write("<"+str(node.label)) #+">") type = "+str(node.type)+" identifier: " + str(node.identifier))
		except Exception:
			#sys.stdout.write("<ERR")#>") type = "+str(node.type)+" identifier: " + str(node.identifier))
			sys.stdout.write("<"+str(node.label.encode('ascii','ignore')))#sys.stdout.write("<!--")#>") type = "+str(node.type)+" identifier: " + str(node.identifier))
		#sys.stdout.write("  attributes = \n")
		for item in node.attrs.keys():
			try:
				if (item != "TextOrElement"):
					sys.stdout.write("  "+str(item.encode('ascii','ignore'))+"= \""+str(node.attrs[item].encode('ascii','ignore'))+"\"")
			except Exception:
					sys.stdout.write("  "+str(item)+"= \""+str(node.attrs[item])+"\"")#sys.stdout.write("  ")
		sys.stdout.write(" kinga_id=\""+str(node.identifier)+"\" >\n")
	elif (node.type == "text" and len(node.label) > 0):	#prints out text nodes
		try:
			sys.stdout.write(str(node.label)+" kinga_id["+str(node.identifier)+"] "+"\n")
		except Exception:
			sys.stdout.write(str(node.label.encode('ascii','ignore'))+" kinga_id["+str(node.identifier)+"] "+"\n")#sys.stdout.write("ERR kinga_id["+str(node.identifier)+"] "+"\n")
		
	for item in node.children:
		printNodeH(item,count+1)
	if (node.type != "text" and len(node.label) > 0 and node.label != "no_label"):
		try:
			sys.stdout.write("</"+str(node.label)+">\n")	
		except Exception:
			sys.stdout.write("</"+str(node.label.encode('ascii','ignore'))+">\n")#sys.stdout.write("-->\n")	
	
	
	#sys.stdout.write("\n");
		
def printNode (node):
	printNodeH(node,0)
	
# used to check if the node only contains whitespace
def checkWhiteSpace(node):
	if(node.nodeType == 3):
		temp = re.search('\S+', node.nodeValue)
		if(temp == None):
			return True
	return False


# creates a unique index for each node in the tree
def getIndex():
	global indexId
	indexId = indexId + 1
	return indexId

# indexes the nodes in a level order sequence
def index(node):
	global maxDepth
	queue = []
	node.depth = 0
	queue.append(node)
	while(len(queue) != 0):
		element = queue.pop(0)
		element.identifier = getIndex()
		if(element.depth != 0):
			element.depth = element.parent.depth + 1
		if(element.depth > maxDepth):
			maxDepth = element.depth
		for child in element.children:
			queue.append(child)

# used to determine whether (x, _) is an element of M
def xInM(x, M):
	for item in M:
		if(item[0] == x):
			return True
	return False

# used to determine whether (_, y) is an element of M
def yInM(y, M):
	#sys.stderr.write("y: " + str(y.identifier) + " M:")
	for item in M:
		#sys.stderr.write(str(item[1].identifier)+"\n")
		if(item[1] == y):
			return True
	#sys.stderr.write("\n")
	return False

# used to retrieve the x-value when given the corresponding y-value
def findX(y, M):
	for item in M:
		if(item[1] == y):
			return item[0]
	return None

#input is nodes x & y, checks to see if they have the same number and type of attributes
def matchAttr(x,y):

	if (len(x.attrs.keys()) != len(y.attrs.keys())):
		return False
	else:
		for item in x.attrs.keys():
			if (not item in y.attrs):
				return False
			if (x.attrs[item] != y.attrs[item]):
				return False
	return True

# input is nodes x & y, and map M, output is map M'
def matchFragment(x, y, M, M2):
	global Mglobal
	global matchAttrG
	#sys.stderr.write("trying to match " + str(x.label) +" and " + str(y.label)+"\n")
	#sys.stderr.write("with            " + str(x.identifier) +" and " + str(y.identifier)+"\n")
	#for item in Mglobal:
	#	sys.stderr.write("Mglobal: " + str(item[0].identifier)+"," + str(item[1].identifier) + "\n")
	attr = matchAttr(x,y)
	#sys.stderr.write("matchAttrG: " + str(matchAttrG) + " (matchAttrG == False): "+str((matchAttrG == "False")) + " attr: " + str(attr) + "\n");
	if (matchAttrG == "False"):
		attr = True
	#sys.stderr.write("attr: " + str(attr) + "\n");
	if(not xInM(x,M) and not yInM(y,M) and (x.type == y.type) and (x.label == y.label) and attr and not xInM(x,Mglobal) and not yInM(y,Mglobal)):
		#try:
		#	sys.stderr.write("matched "+str(x.label)+", mapping ["+str(x.identifier)+","+str(y.identifier)+"]\n")
		#except Exception:
		#	sys.stderr.write("matched ERR, mapping ["+str(x.identifier)+","+str(y.identifier)+"]\n")
		M2 = M2.union([(x,y)])
		Mglobal = Mglobal.union(M2)
		if(len(x.children) > len(y.children)):
			#sys.stderr.write("x has more children\n")
			i = 0
			while(len(y.children) > i):
				M2 = M2.union(matchFragment(x.children[i], y.children[i], M, M2))
				i = i + 1
		else:
			#sys.stderr.write("y has more or equal children, with "+str(len(y.children))+"children for y\n")
			i = 0
			while(len(x.children) > i):
				m = (matchFragment(x.children[i], y.children[i], M, M2))
				j = i
				while (len(m) == 0 and j < len(y.children)):
					#sys.stderr.write("failed current child on " + str(x.label)+"\n")
					j = j + 1
					if (j < len(y.children)):
						m = (matchFragment(x.children[i], y.children[j], M, M2))
				M2 = M2.union(m)
				i = i + 1
	else:
		#try:
		#	sys.stderr.write("could not match (" + str(x.type)+","+str(y.type)+") and ("+str(x.label)+","+str(y.label)+") and ("+str(x.identifier)+","+str(y.identifier)+"\n")  
		#except Exception:
		#	sys.stderr.write("could not match (" + str(x.type)+","+str(y.type)+") and (ERR,ERR) and ("+str(x.identifier)+","+str(y.identifier)+"\n")  
		M2 = set()
	return M2

# input is two trees, output is a map
# computes the mapping from T1 to T2
# equality is when two nodes have the same type and label
def mapping(T1, T2, M):
	queue = []
	queue.append(T1)
	while(len(queue) != 0):
		x = queue.pop(0)
		for child in x.children:
			queue.append(child)
		if(xInM(x,M)):
			continue
		y = []
		yQueue = []
		yQueue.append(T2)
		while(len(yQueue) != 0):
			element = yQueue.pop(0)
			for child in element.children:
				yQueue.append(child)
			if(element.label == x.label and element.type == x.type):
				y.append(element)
				#sys.stderr.write("appended: " + str(element.type)+"\n")
		
		M3 = set()
		
		i = 0
		while(i < len(y)):
			if(not yInM(y[i],M)):	
				M2 = set()
				M2 = matchFragment(x, y[i], M, M2)
				if (len(M2) > len(M3)):
					M3 = M2
			i = i + 1
		M = M.union(M3)
	return M
			
# used to convert the nodes from the XML parser into my nodes, this is not necessary but it requires changing less code
def convertNodeXML(node, walker, treeId):
	type = ""
	label = ""
	if(walker.currentNode.nodeType == 1): # 1 is used by the parser to denote an element node
		type = "element"
		label = walker.currentNode.nodeName
	elif(walker.currentNode.nodeType == 3): # 3 is used by the parser to denote a text node
		type = "text"
		label = walker.currentNode.nodeValue
	parent = None
	position = 1
	attrs = walker.currentNode.attributes #NamedNodeMap mapping attribute names to Attribute instances
	dictAttrs = {"TextOrElement":label}
	if not(attrs == None):
		for attrName in attrs.keys():
			attrNode = attrs.get(attrName)
			attrValue = attrNode.nodeValue
			dictAttrs[attrName[1].lower()] = attrValue
			#sys.stdout.write("attrName: " + str(attrName[1])+ "\n")
			#sys.stdout.write("attribute: " + str(attrValue) + "\n")

	#for i in dictAttrs.keys():
	#	sys.stdout.write(str(i)+"->"+str(dictAttrs[i])+"\n")
	
	#adds the labels seen to the global hashtable for all labels seen for each file
	if(walker.currentNode.nodeType == 1):
		if treeId == "1":
			elementTypes1[label] = "-"
		else:
			elementTypes2[label] = "-"
	
	newNode = Node(label, type, parent, position, dictAttrs)
	for child in walker.currentNode.childNodes:
		next = walker.nextNode()
		if next is None: break
		if(not checkWhiteSpace(child)):
			n = convertNodeXML(child, walker,treeId)
			n.parent = newNode
			newNode.children.append(n)
	i = 1
	for child in newNode.children:
		child.position = i
		i = i + 1
	return newNode

# used to convert the nodes from the HTML parser into my nodes
def convertNodeHTML(node):
	
	type = ""
	label = ""
	if(node.nodeType == 1): # 1 is used by the parser to denote an element node
		type = "element"
		label = node.nodeName
	elif(node.nodeType == 3): # 3 is used by the parser to denote a text node
		type = "text"
		label = node.nodeValue
	elif(node.nodeType == 9):
		type = "document"
		label = "no_label"
	#else:
	#	type = "element"
	#	label = node.nodeName
	parent = None
	position = 1

	attrs = node.attributes #NamedNodeMap mapping attribute names to Attribute instances
	dictAttrs = {"TextOrElement":label}
	if not(attrs == None):
		for attrName in attrs.keys():
			attrNode = attrs.get(attrName)
			attrValue = attrNode.nodeValue
			dictAttrs[attrName] = attrValue
			#sys.stdout.write("attrName: " + str(attrName[1])+ "\n")
			#sys.stdout.write("attribute: " + str(attrValue) + "\n")

	#for i in dictAttrs.keys():
	#	sys.stdout.write(str(i)+"->"+str(dictAttrs[i])+"\n")

	newNode = Node(label, type, parent, position, dictAttrs)
	for child in node.childNodes:
		if(not checkWhiteSpace(child)):
			n = convertNodeHTML(child)
			n.parent = newNode
			newNode.children.append(n)
	i = 1
	for child in newNode.children:
		child.position = i
		i = i + 1
	return newNode

def upon_alarm():
        sys.stdout.write("crashed")
	sys.stdout.flush()
	os._exit(1)


# main method
def main():
	global fileType
	global Mglobal
	global matchAttrG
	if(True):
		alarm = threading.Timer(500.0, upon_alarm)
        	alarm.start()
		# the files to read from
		f1 = "faults/"+ sys.argv[1]
		f2 = "faults/"+ sys.argv[2]
		file1 = open(f1)
		file2 = open(f2)
		root1 = None
		root2 = None
		if(f1.endswith(".xml") and f2.endswith(".xml")):
			fileType = "xml"
			# create the xml parser
			reader = Sax2.Reader()	
			# parse the file and store it
			tree1 = reader.fromUri(f1)
			tree2 = reader.fromUri(f2)
			walker1 = tree1.createTreeWalker(tree1.documentElement, NodeFilter.SHOW_ALL, None, 0)
			walker2 = tree2.createTreeWalker(tree2.documentElement, NodeFilter.SHOW_ALL, None, 0)
			root1 = convertNodeXML(tree1, walker1, "1")
			root2 = convertNodeXML(tree2, walker2, "2")
			index(root1)
			index(root2)
		else:
			fileType = "html"
			#create the html parser
			parser = html5lib.HTMLParser(tree=treebuilders.getTreeBuilder("dom"))
			# parse the file and store it
			tree1 = parser.parse(file1)
			tree2 = parser.parse(file2)
			root1 = convertNodeHTML(tree1)
			root2 = convertNodeHTML(tree2)
			index(root1)
			index(root2)
		#else:
		#	sys.stdout.write("Make sure both files have the same extension, either .html or .xml. Program will exit.\n")
		#	sys.stdout.flush()
		#	os._exit(1)
	        M = set()
		matchAttrG = sys.argv[3]
	        result = mapping(root1, root2, M)
		sys.stderr.write("done mapping\n");
		printNode(root1)
		sys.stderr.write("done print 1\n")
		sys.stdout.write("###################\n")
		printNode(root2)
		sys.stderr.write("done print 1\n")
		sys.stdout.write("###################\n")
		for item in Mglobal:
			sys.stdout.write("["+str(item[0].identifier)+","+str(item[1].identifier)+"]\n")
		sys.stderr.write("DONE!\n")
		alarm.cancel()	

if __name__=="__main__":
	main()

